package de.unibi.citec.clf.bonsai.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;
import java.util.ArrayList;

import de.unibi.citec.clf.bonsai.core.Actuator;

/**
 * Interface for actuators controlling a speech synthesizer.
 *
 * @author jwienke
 * @author arothert
 */
public abstract class SpeechActuator implements Actuator {

    /**
     * Say method with the ability to chose between synchronous and asynchronous
     * speech synthesis.
     *
     * @param text
     *            text to say
     * @param async
     *            if <code>true</code>, the method call will return as fast as
     *            possible, blocks otherwise
     * @throws IOException
     *             communication error
     */
    public abstract void say(String text, boolean async) throws IOException;

    /**
     * Synchronous say method. Returns after the complete text was said.
     *
     * @param text
     *            text to say
     * @throws IOException
     *             communication error
     */
    public abstract void say(String text) throws IOException;

    /**
     * For self-written prosody for the whole text.
     * @param accented_text
     * @throws IOException
     */
    public abstract void sayAccentuated(String accented_text) throws IOException;

    /**
     * For normal text, maybe accentuations within the text.
     * Additionally a configuration file path to set the overall prosody of a text must be given.
     * @param accented_text
     * @param prosodyConfig
     * @throws IOException
     */
    public abstract void sayAccentuated(String accented_text, String prosodyConfig) throws IOException;

    /**
     * For self-written prosody for the whole text.
     * @param accented_text
     *          text to say, might invole specific accentuations of parts
     *          of the text (MARYXML - prosody - style).
     * @param async
     *          if <code>true</code>, wait until the complete text was said,
     *          else return as fast as possible from this method call
     * @throws IOException
     */
    public abstract void sayAccentuated(String accented_text, boolean async) throws IOException;

    /**
     * For normal text, maybe accentuations within the text.
     * Additionally a configuration file path to set the overall prosody of a text must be given.
     *
     * @param accented_text
     *          text to say, might invole specific accentuations of parts
     *          of the text (MARYXML - prosody - style).
     * @param async
     *          if <code>true</code>, wait until the complete text was said,
     *          else return as fast as possible from this method call
     * @param prosodyConfig
     *          path and file for overall prosody setting of the accented_text
     * @throws IOException
     */
    public abstract void sayAccentuated(String accented_text, boolean async, String prosodyConfig) throws IOException;
    
    protected static ArrayList<SpeechActuatorListener> listeners = new ArrayList<SpeechActuatorListener>();
    
    
    public static void addListener(SpeechActuatorListener listener) {
		listeners.add(listener);
	}

	public static void removeListener(SpeechActuatorListener listener) {
		listeners.remove(listener);
	}
	
	public abstract void setIsrActuator(IsrActuator isrAct);
}
