package de.unibi.citec.clf.bonsai.core.configuration;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import nu.xom.Builder;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.Node;
import nu.xom.Nodes;
import nu.xom.ParsingException;
import nu.xom.ValidityException;

import org.apache.log4j.Logger;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

/**
 * Parser for XML configuration files. A schema is installed with BonSAI
 * specifying the expected format.
 * 
 * <p>
 * The expected format defines, that all sensors must be defined before any
 * actuator is specified.
 * </p>
 * 
 * @author jwienke
 */
public class XmlConfigurationParser implements ConfigurationParser {

	private Logger logger = Logger.getLogger(getClass());
	private static final String ROOT_ELEMENT_NAME = "BonsaiConfiguration";
	private static final String FACTORY_OPTIONS_ELEMENT_NAME = "FactoryOptions";
	private static final String MEMORY_ELEMENT_NAME = "WorkingMemory";
	private static final String SLOT_ELEMENT_NAME = "Slot";
	private static final String SLOTS_ELEMENT_NAME = "Slots";
	private static final String SLOT_CLASS_ARG = "slotClass";
	private static final String OPTION_ELEMENT_NAME = "Option";
	private static final String SENSOR_ELEMENT_NAME = "Sensor";
	private static final String OPTIONS_ELEMENT_NAME = "Options";
	private static final String FACTORY_CLASS_ARG = "factoryClass";
	private static final String MEMORY_CLASS_ARG = "workingMemoryClass";
	private static final String OPTION_KEY_ARG = "key";
	private static final String SENSOR_KEY_ARG = "key";
	private static final String MEMORY_KEY_ARG = "key";
	private static final String DATA_TYPE_CLASS_ARG = "dataTypeClass";
	private static final String LIST_TYPE_CLASS_ARG = "listTypeClass";
	private static final String SENSOR_CLASS_ARG = "sensorClass";
	private static final String ACTUATOR_ELEMENT_NAME = "Actuator";
	private static final String ACTUATOR_KEY_ARG = "key";
	private static final String ACTUATOR_CLASS_ARG = "actuatorClass";
	private static final String ACTUATOR_INTERFACE_ARG = "actuatorInterface";
	private Map<String, ConfiguredCoreObjectFactory> factoriesByClass = new HashMap<String, ConfiguredCoreObjectFactory>();
	private Map<String, ConfiguredWorkingMemory> memoriesByKey = new HashMap<String, ConfiguredWorkingMemory>();
	private Map<String, ConfiguredSensor> sensorsByKey = new HashMap<String, ConfiguredSensor>();
	private Map<String, ConfiguredActuator> actuatorsByKey = new HashMap<String, ConfiguredActuator>();
	private boolean parsingCompleted = false;

	/**
	 * Constructor.
	 */
	public XmlConfigurationParser() {
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Set<ConfiguredCoreObjectFactory> getConfiguredFactories()
			throws IllegalStateException {
		if (!parsingCompleted) {
			throw new IllegalStateException(
					"Parsing was not invoked or failed. "
							+ "No results can be fetched.");
		}
		return new HashSet<ConfiguredCoreObjectFactory>(
				factoriesByClass.values());
	}

	@Override
	public Set<ConfiguredWorkingMemory> getConfiguredWorkingMemories()
			throws IllegalStateException {
		if (!parsingCompleted) {
			throw new IllegalStateException(
					"Parsing was not invoked or failed. "
							+ "No results can be fetched.");
		}
		return new HashSet<ConfiguredWorkingMemory>(memoriesByKey.values());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Set<ConfiguredSensor> getConfiguredSensors()
			throws IllegalStateException {
		if (!parsingCompleted) {
			throw new IllegalStateException(
					"Parsing was not invoked or failed. "
							+ "No results can be fetched.");
		}
		return new HashSet<ConfiguredSensor>(sensorsByKey.values());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Set<ConfiguredActuator> getConfiguredActuators()
			throws IllegalStateException {
		if (!parsingCompleted) {
			throw new IllegalStateException(
					"Parsing was not invoked or failed. "
							+ "No results can be fetched.");
		}
		return new HashSet<ConfiguredActuator>(actuatorsByKey.values());
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void parseFile(URI configurationUri) throws IOException,
			ParseException, IllegalStateException {

		
		File configurationFile;
		try {
			configurationFile = new File(configurationUri.getPath());
			if (!configurationFile.exists() || !configurationFile.canRead()) {
				throw new IOException(
						"Given configuration file does not exist: "
								+ configurationFile);
			}
		} catch (IllegalArgumentException e) {
			throw new IOException("IOException while reading config file: "
					+ configurationUri + ": " + e.getMessage(), e);
		}

		if (parsingCompleted) {
			logger.warn("Parser already parsed a file");
		}

		try {
			XMLReader xerces = XMLReaderFactory
					.createXMLReader("org.apache.xerces.parsers.SAXParser");
			xerces.setFeature(
					"http://apache.org/xml/features/validation/schema", true);

			Builder parser = new Builder(xerces, true);

			Document doc = parser.build(configurationFile.getAbsoluteFile());

			parseFactorySettings(doc);
			parseSensors(doc);
			parseActuators(doc);
			parseWorkingMemories(doc);

			parsingCompleted = true;

		} catch (ValidityException e) {
			throw new ParseException(e);
		} catch (ParsingException e) {
			throw new ParseException(e);
		} catch (SAXException e) {
			throw new ParseException("Maybe xerces SAXParser missing?", e);
		}
	}

	private void parseFactorySettings(Document doc) throws ParseException {

		logger.debug("Parsing factory settings");

		Nodes factoryOptions = doc.query("/" + ROOT_ELEMENT_NAME + "/"
				+ FACTORY_OPTIONS_ELEMENT_NAME);
		for (int i = 0; i < factoryOptions.size(); i++) {

			Node factoryOptionNode = factoryOptions.get(i);
			logger.debug("Possible " + FACTORY_OPTIONS_ELEMENT_NAME + ": "
					+ factoryOptionNode.toXML());
			if (!(factoryOptionNode instanceof Element)) {
				logger.warn("Skipping because no "
						+ Element.class.getSimpleName());
				continue;
			}
			Element factoryOption = (Element) factoryOptionNode;

			ConfiguredCoreObjectFactory factory = new ConfiguredCoreObjectFactory();

			// class name
			String factoryClassName = factoryOption
					.getAttributeValue(FACTORY_CLASS_ARG);
			factory.setFactoryClassName(getTrimmedValue(factoryClassName));

			// options
			Map<String, String> options = parseOptions(factoryOption);
			factory.setFactroyOptions(options);

			// check that a configuration for this element doesn't already exist
			if (factoriesByClass.containsKey(factory.getFactoryClassName())) {
				logger.warn("There is already a configuration for factory class "
						+ factory.getFactoryClassName());
			}
			factoriesByClass.put(factory.getFactoryClassName(), factory);

		}

	}

	private void parseWorkingMemories(Document doc) throws ParseException {

		logger.debug("Parsing working memory settings");

		Nodes memoryOptions = doc.query("/" + ROOT_ELEMENT_NAME + "/"
				+ MEMORY_ELEMENT_NAME);
		for (int i = 0; i < memoryOptions.size(); i++) {

			Node memoryNode = memoryOptions.get(i);
			logger.debug("Possible " + MEMORY_ELEMENT_NAME + ": "
					+ memoryNode.toXML());
			if (!(memoryNode instanceof Element)) {
				logger.warn("Skipping because no "
						+ Element.class.getSimpleName());
				continue;
			}
			Element memory = (Element) memoryNode;

			ConfiguredWorkingMemory confMemory = new ConfiguredWorkingMemory();

			// key
			String key = memory.getAttributeValue(MEMORY_KEY_ARG);
			confMemory.setKey(getTrimmedValue(key));

			// factory
			String factory = memory.getAttributeValue(FACTORY_CLASS_ARG);
			confMemory.setFactoryClassName(getTrimmedValue(factory));

			// class name
			String className = memory.getAttributeValue(MEMORY_CLASS_ARG);
			confMemory.setMemoryClassName(getTrimmedValue(className));

			// options
			Element optionsElement = memory
					.getFirstChildElement(OPTIONS_ELEMENT_NAME);
			if (optionsElement == null) {
				throw new ParseException(MEMORY_ELEMENT_NAME
						+ " does not contain an " + OPTIONS_ELEMENT_NAME + ": "
						+ memory.toXML());
			}
			Map<String, String> options = parseOptions(optionsElement);
			confMemory.setOptions(options);

			// slots
			Element slotsElement = memory
					.getFirstChildElement(SLOTS_ELEMENT_NAME);
			if (slotsElement == null) {
				throw new ParseException(MEMORY_ELEMENT_NAME
						+ " does not contain an " + SLOTS_ELEMENT_NAME + ": "
						+ memory.toXML());
			}
			Set<ConfiguredMemorySlot> slots = parseSlots(slotsElement);
			confMemory.setMemorySlots(slots);

			// check that a configuration for this element doesn't already exist
			if (memoriesByKey.containsKey(confMemory.getKey())) {
				logger.warn("There is already a configuration for working memory key "
						+ confMemory.getKey());
			}
			memoriesByKey.put(confMemory.getKey(), confMemory);

		}

	}

	private void parseSensors(Document doc) throws ParseException {

		logger.debug("Parsing sensors");

		Nodes sensorNodes = doc.query("/" + ROOT_ELEMENT_NAME + "/"
				+ SENSOR_ELEMENT_NAME);
		for (int i = 0; i < sensorNodes.size(); i++) {

			Node sensorNode = sensorNodes.get(i);
			logger.debug("Possible " + SENSOR_ELEMENT_NAME + ": "
					+ sensorNode.toXML());
			if (!(sensorNode instanceof Element)) {
				logger.warn("Skipping because no "
						+ Element.class.getSimpleName());
				continue;
			}
			Element sensor = (Element) sensorNode;

			ConfiguredSensor configuredSensor = new ConfiguredSensor();

			// key
			String key = sensor.getAttributeValue(SENSOR_KEY_ARG);
			configuredSensor.setKey(getTrimmedValue(key));

			// data type class
			String dataTypeClass = sensor
					.getAttributeValue(DATA_TYPE_CLASS_ARG);
			configuredSensor.setDataType(getTrimmedValue(dataTypeClass));

			// data type class
			String listTypeClass = sensor
					.getAttributeValue(LIST_TYPE_CLASS_ARG);
			if (listTypeClass != null) {
				logger.debug("This sensor expects a list");
				configuredSensor.setListType(getTrimmedValue(listTypeClass));
			}

			// factory class
			String factoryClass = sensor.getAttributeValue(FACTORY_CLASS_ARG);
			configuredSensor.setFactoryClassName(getTrimmedValue(factoryClass));

			// sensor class
			String sensorClass = sensor.getAttributeValue(SENSOR_CLASS_ARG);
			configuredSensor.setSensorClassName(getTrimmedValue(sensorClass));

			// options
			Element optionElement = sensor
					.getFirstChildElement(OPTIONS_ELEMENT_NAME);
			if (optionElement == null) {
				throw new ParseException(SENSOR_ELEMENT_NAME
						+ " does not contain an " + OPTIONS_ELEMENT_NAME + ": "
						+ sensor.toXML());
			}
			configuredSensor.setOptions(parseOptions(optionElement));

			// check that a sensor with this key doesn't already exist
			if (sensorsByKey.containsKey(configuredSensor.getKey())) {
				logger.warn("There is already a sensor for key "
						+ configuredSensor.getKey());
			}
			sensorsByKey.put(configuredSensor.getKey(), configuredSensor);

		}

	}

	private void parseActuators(Document doc) throws ParseException {

		logger.debug("Parsing actuators");

		Nodes actuatorNodes = doc.query("/" + ROOT_ELEMENT_NAME + "/"
				+ ACTUATOR_ELEMENT_NAME);
		for (int i = 0; i < actuatorNodes.size(); i++) {

			Node actuatorNode = actuatorNodes.get(i);
			logger.debug("Possible " + ACTUATOR_ELEMENT_NAME + ": "
					+ actuatorNode.toXML());
			if (!(actuatorNode instanceof Element)) {
				logger.warn("Skipping because no "
						+ Element.class.getSimpleName());
				continue;
			}
			Element sensor = (Element) actuatorNode;

			ConfiguredActuator configuredActuator = new ConfiguredActuator();

			// key
			String key = sensor.getAttributeValue(ACTUATOR_KEY_ARG);
			configuredActuator.setKey(getTrimmedValue(key));

			// actuator class
			String actuatorClass = sensor.getAttributeValue(ACTUATOR_CLASS_ARG);
			configuredActuator
					.setActuatorClassName(getTrimmedValue(actuatorClass));

			// factory class
			String factoryClass = sensor.getAttributeValue(FACTORY_CLASS_ARG);
			configuredActuator
					.setFactoryClassName(getTrimmedValue(factoryClass));

			// interface
			String interfaceClass = sensor
					.getAttributeValue(ACTUATOR_INTERFACE_ARG);
			configuredActuator
					.setInterfaceClassName(getTrimmedValue(interfaceClass));

			// options
			Element optionElement = sensor
					.getFirstChildElement(OPTIONS_ELEMENT_NAME);
			if (optionElement == null) {
				throw new ParseException(SENSOR_ELEMENT_NAME
						+ " does not contain an " + OPTIONS_ELEMENT_NAME + ": "
						+ sensor.toXML());
			}
			configuredActuator.setOptions(parseOptions(optionElement));

			// check that a sensor with this key doesn't already exist
			if (actuatorsByKey.containsKey(configuredActuator.getKey())) {
				logger.warn("There is already a actuator for key "
						+ configuredActuator.getKey());
			}
			actuatorsByKey.put(configuredActuator.getKey(), configuredActuator);

		}

	}

	private Map<String, String> parseOptions(Element optionRoot)
			throws ParseException {

		logger.debug("Parsing options in root: " + optionRoot.toXML());

		Map<String, String> options = new HashMap<String, String>();

		Nodes optionNodes = optionRoot.query(OPTION_ELEMENT_NAME);
		for (int i = 0; i < optionNodes.size(); i++) {

			Node optionNode = optionNodes.get(i);
			logger.debug("Possible " + OPTION_ELEMENT_NAME + ": "
					+ optionNode.toXML());
			if (!(optionNode instanceof Element)) {
				logger.warn("Skipping because no "
						+ Element.class.getSimpleName());
				continue;
			}
			Element option = (Element) optionNode;

			String key = getTrimmedValue(option
					.getAttributeValue(OPTION_KEY_ARG));
			String value = getTrimmedValue(option.getValue());

			options.put(key, value);

		}

		return options;

	}

	private Set<ConfiguredMemorySlot> parseSlots(Element slotRoot)
			throws ParseException {

		logger.debug("Parsing slots in root: " + slotRoot.toXML());

		Set<ConfiguredMemorySlot> slots = new HashSet<ConfiguredMemorySlot>();

		Nodes slotNodes = slotRoot.query(SLOT_ELEMENT_NAME);
		for (int i = 0; i < slotNodes.size(); i++) {

			Node slotNode = slotNodes.get(i);
			logger.debug("Possible " + SLOT_ELEMENT_NAME + ": "
					+ slotNode.toXML());
			if (!(slotNode instanceof Element)) {
				logger.warn("Skipping because no "
						+ Element.class.getSimpleName());
				continue;
			}
			Element slot = (Element) slotNode;

			ConfiguredMemorySlot confSlot = new ConfiguredMemorySlot();
			confSlot.setDataType(getTrimmedValue(slot
					.getAttributeValue(DATA_TYPE_CLASS_ARG)));
			confSlot.setSlotClassName(getTrimmedValue(slot
					.getAttributeValue(SLOT_CLASS_ARG)));

			slots.add(confSlot);
		}

		return slots;

	}

	/**
	 * Returns a trimmed value of the given string and throws
	 * {@link ParseException} for every error that can occur. This may be a
	 * <code>null</code> value or an empty string.
	 * 
	 * @param rawValue
	 *            value to trim, may be <code>null</code>
	 * @return trimmed string value, not <code>null</code>
	 * @throws ParseException
	 *             empty string or <code>null</code> value
	 */
	private String getTrimmedValue(String rawValue) throws ParseException {

		if (rawValue == null) {
			throw new ParseException("Null value");
		}
		String finalValue = rawValue.trim();
		if (finalValue.isEmpty()) {
			throw new ParseException("Empty value");
		}

		return finalValue;

	}
}
