package de.unibi.citec.clf.bonsai.core;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.airobots.btl.List;
import de.unibi.airobots.btl.Type;
import de.unibi.airobots.btl.data.geometry.Point3D;

import de.unibi.airobots.btl.data.navigation.PositionData;
import de.unibi.citec.clf.bonsai.core.BonsaiManager.ConfigurationException;
import de.unibi.citec.clf.bonsai.core.configuration.ConfigurationParser;
import de.unibi.citec.clf.bonsai.core.configuration.ConfiguredActuator;
import de.unibi.citec.clf.bonsai.core.configuration.ConfiguredCoreObjectFactory;
import de.unibi.citec.clf.bonsai.core.configuration.ConfiguredMemorySlot;
import de.unibi.citec.clf.bonsai.core.configuration.ConfiguredSensor;
import de.unibi.citec.clf.bonsai.core.configuration.ConfiguredWorkingMemory;

import java.io.IOException;
import java.net.URI;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Logger;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Test cases for the {@link BonsaiManager}.
 * 
 * @author jwienke
 */
public class BonsaiManagerTest {

	/**
	 * A simple dummy configurator for the unit test.
	 * 
	 * @author jwienke
	 */
	private class DummyConfigurator implements ConfigurationParser {

		private boolean illegalState = false;
		private boolean ioException = false;
		private boolean parserError = false;
		private boolean illegalParseState = false;

		private Set<ConfiguredActuator> configuredActuators = new HashSet<ConfiguredActuator>();
		private Set<ConfiguredCoreObjectFactory> configuredFactories = new HashSet<ConfiguredCoreObjectFactory>();
		private Set<ConfiguredWorkingMemory> configuredMemories = new HashSet<ConfiguredWorkingMemory>();
		private Set<ConfiguredSensor> configuredSensors = new HashSet<ConfiguredSensor>();

		@Override
		public Set<ConfiguredActuator> getConfiguredActuators()
				throws IllegalStateException {
			if (illegalState) {
				throw new IllegalStateException();
			}
			return configuredActuators;
		}

		@Override
		public Set<ConfiguredCoreObjectFactory> getConfiguredFactories()
				throws IllegalStateException {
			if (illegalState) {
				throw new IllegalStateException();
			}
			return configuredFactories;
		}

		@Override
		public Set<ConfiguredSensor> getConfiguredSensors()
				throws IllegalStateException {
			if (illegalState) {
				throw new IllegalStateException();
			}
			return configuredSensors;
		}

		@Override
		public void parseFile(URI configurationUrl) throws IOException,
				ParseException, IllegalStateException {
			if (ioException) {
				throw new IOException();
			}
			if (parserError) {
				throw new ParseException("");
			}
			if (illegalParseState) {
				throw new IllegalStateException();
			}
		}

		@Override
		public Set<ConfiguredWorkingMemory> getConfiguredWorkingMemories()
				throws IllegalStateException {
			if (illegalState) {
				throw new IllegalStateException();
			}
			return configuredMemories;
		}

	}

	private DummyConfigurator configurator;
	private static Logger logger;

	/**
	 * Setup logging.
	 * 
	 * @throws Exception
	 *             error
	 */
	@BeforeClass
	public static void setUpBeforeClass() throws Exception {
		BasicConfigurator.configure();
		logger = Logger.getLogger(BonsaiManagerTest.class);
	}

	/**
	 * Setup method constructing a new "singleton" instance of
	 * {@link BonsaiManager}.
	 */
	@Before
	public void setUp() {
		BonsaiManager.testConstructNewInstance();
		configurator = new DummyConfigurator();
	}

	@Test(expected = ConfigurationException.class)
	public void parseErrorIOException() throws Exception {
		configurator.ioException = true;
		BonsaiManager.getInstance().configure("dummy", configurator);
	}

	@Test(expected = ConfigurationException.class)
	public void parseErrorParseException() throws Exception {
		configurator.parserError = true;
		BonsaiManager.getInstance().configure("dummy", configurator);
	}

	@Test(expected = ConfigurationException.class)
	public void illegealState() throws Exception {
		configurator.illegalParseState = true;
		BonsaiManager.getInstance().configure("dummy", configurator);
	}

	@Test(expected = ConfigurationException.class)
	public void unknownFactoryClass() throws Exception {

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName("I.dont.know.You");
		f.setFactroyOptions(new HashMap<String, String>());
		configurator.configuredFactories.add(f);

		BonsaiManager.getInstance().configure("dummy", configurator);

	}

	@Test(expected = ConfigurationException.class)
	public void noFactoryClass() throws Exception {

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(String.class.getName());
		f.setFactroyOptions(new HashMap<String, String>());
		configurator.configuredFactories.add(f);

		BonsaiManager.getInstance().configure("dummy", configurator);

	}

	@Test(expected = ConfigurationException.class)
	public void duplicateFactoryClass() throws Exception {

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(String.class.getName());
		f.setFactroyOptions(new HashMap<String, String>());
		configurator.configuredFactories.add(f);
		ConfiguredCoreObjectFactory f2 = new ConfiguredCoreObjectFactory();
		f2.setFactoryClassName(String.class.getName());
		f2.setFactroyOptions(new HashMap<String, String>());
		configurator.configuredFactories.add(f2);

		BonsaiManager.getInstance().configure("dummy", configurator);

	}

	@Test
	public void workingFactory() throws Exception {

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		f.setFactroyOptions(new HashMap<String, String>());
		configurator.configuredFactories.add(f);

		BonsaiManager.getInstance().configure("dummy", configurator);

		assertEquals(1, BonsaiManager.getInstance()
				.getConfiguredFactoryClasses().size());
		assertTrue(BonsaiManager.getInstance().getConfiguredFactoryClasses()
				.contains(TestFactory.class));

	}

	@Test(expected = ConfigurationException.class)
	public void factoryConfigurationError() throws Exception {

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		Map<String, String> options = new HashMap<String, String>();
		options.put(TestFactory.FAIL_KEY, "i will fail misserably");
		f.setFactroyOptions(options);
		configurator.configuredFactories.add(f);

		BonsaiManager.getInstance().configure("dummy", configurator);

	}

	@Test(expected = ConfigurationException.class)
	public void duplicateKeyError() throws Exception {

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		Map<String, String> options = new HashMap<String, String>();
		f.setFactroyOptions(options);
		configurator.configuredFactories.add(f);

		ConfiguredSensor s1 = new ConfiguredSensor();
		s1.setKey("duplicate");
		s1.setDataType(String.class.getName());
		s1.setFactoryClassName(TestFactory.class.getName());
		s1.setSensorClassName(TestSensor.class.getName());
		s1.setOptions(new HashMap<String, String>());
		configurator.configuredSensors.add(s1);

		ConfiguredSensor s2 = new ConfiguredSensor();
		s2.setKey(s1.getKey());
		s2.setDataType(String.class.getName());
		s2.setFactoryClassName(TestFactory.class.getName());
		s2.setSensorClassName(TestSensor.class.getName());
		s2.setOptions(new HashMap<String, String>());
		configurator.configuredSensors.add(s2);

		BonsaiManager.getInstance().configure("dummy", configurator);

	}

	@Test
	public void correctSensorCreation() throws Exception {

		final String key = "testKey";

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		Map<String, String> options = new HashMap<String, String>();
		f.setFactroyOptions(options);
		configurator.configuredFactories.add(f);

		ConfiguredSensor s1 = new ConfiguredSensor();
		s1.setKey(key);
		s1.setDataType(String.class.getName());
		s1.setSensorClassName(TestSensor.class.getName());
		s1.setFactoryClassName(TestFactory.class.getName());
		s1.setOptions(new HashMap<String, String>());
		configurator.configuredSensors.add(s1);

		BonsaiManager.getInstance().configure("dummy", configurator);

		assertTrue(BonsaiManager.getInstance().canCreateSensor(key,
				String.class));
		assertFalse(BonsaiManager.getInstance().canCreateSensor(key,
				Integer.class));
		assertFalse(BonsaiManager.getInstance().canCreateSensor("dummyblaa",
				String.class));

		assertNotNull(BonsaiManager.getInstance().createSensor(key,
				String.class));

		assertTrue(BonsaiManager.getInstance().createSensor(key, String.class) instanceof TestSensor<?>);

		// ensure that creating an unknown key-type combination fails
		try {
			BonsaiManager.getInstance().createSensor(key, Integer.class);
			fail("Must throw an exception "
					+ "for an unknown key-type combination.");
		} catch (IllegalArgumentException e) {
		}

		try {
			BonsaiManager.getInstance().createSensor(key + "1", String.class);
			fail("Must throw an exception "
					+ "for an unknown key-type combination.");
		} catch (IllegalArgumentException e) {
		}

	}

	@Test
	public void correctListSensorCreation() throws Exception {

		final String key = "testKey";

		logger.info("  --------------  List Test -------------");

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		Map<String, String> options = new HashMap<String, String>();
		f.setFactroyOptions(options);
		configurator.configuredFactories.add(f);

		ConfiguredSensor s1 = new ConfiguredSensor();
		s1.setKey(key);
		s1.setDataType(PositionData.class.getName());
		s1.setListType(List.class.getName());
		s1.setSensorClassName(TestSensor.class.getName());
		s1.setFactoryClassName(TestFactory.class.getName());
		s1.setOptions(new HashMap<String, String>());
		configurator.configuredSensors.add(s1);

		BonsaiManager.getInstance().configure("dummy", configurator);

		@SuppressWarnings("unchecked")
		Class<List<PositionData>> listType = (Class<List<PositionData>>) new List<PositionData>(
				PositionData.class).getClass();

		assertTrue(BonsaiManager.getInstance().canCreateSensor(key, listType,
				PositionData.class));
		assertFalse(BonsaiManager.getInstance().canCreateSensor(key, listType,
				Integer.class));
		assertFalse(BonsaiManager.getInstance().canCreateSensor("dummyblaa",
				listType, PositionData.class));

		assertNotNull(BonsaiManager.getInstance().createSensor(key, listType,
				PositionData.class));

		assertTrue(BonsaiManager.getInstance().createSensor(key, listType,
				PositionData.class) instanceof TestSensor<?>);

		// ensure that creating an unknown key-type combination fails
		try {
                    Class<List<Point3D>> listType0 =  (Class<List<Point3D>>) List.getListClass(Point3D.class);
			BonsaiManager.getInstance().createSensor(key,listType0
					, Point3D.class);
			fail("Must throw an exception "
					+ "for an unknown key-type combination.");
		} catch (IllegalArgumentException e) {
		}

		try {
			BonsaiManager.getInstance().createSensor(key + "1", listType,
					PositionData.class);
			fail("Must throw an exception "
					+ "for an unknown key-type combination.");
		} catch (IllegalArgumentException e) {
		}

	}

	@Test
	public void correctActuatorCreation() throws Exception {

		final String key = "testKey";

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		Map<String, String> options = new HashMap<String, String>();
		f.setFactroyOptions(options);
		configurator.configuredFactories.add(f);

		ConfiguredActuator actuator = new ConfiguredActuator();
		actuator.setKey(key);
		actuator.setActuatorClassName(FooActuatorImpl.class.getName());
		actuator.setInterfaceClassName(FooActuator.class.getName());
		actuator.setFactoryClassName(TestFactory.class.getName());
		actuator.setOptions(new HashMap<String, String>());
		configurator.configuredActuators.add(actuator);

		BonsaiManager.getInstance().configure("dummy", configurator);

		// check that derived instantiation is possible
		assertTrue(BonsaiManager.getInstance().canCreateActuator(key,
				FooActuator.class));
		assertTrue(BonsaiManager.getInstance().canCreateActuator(key,
				FooActuatorImpl.class));
		assertFalse(BonsaiManager.getInstance().canCreateSensor(key,
				DerivedFooActuatorImpl.class));

		// wrong key should not be possible
		assertFalse(BonsaiManager.getInstance().canCreateActuator(key + "bla",
				FooActuator.class));
		assertFalse(BonsaiManager.getInstance().canCreateActuator(key + "bla",
				FooActuatorImpl.class));
		assertFalse(BonsaiManager.getInstance().canCreateSensor(key + "bla",
				DerivedFooActuatorImpl.class));

		assertNotNull(BonsaiManager.getInstance().createActuator(key,
				FooActuator.class));
		assertNotNull(BonsaiManager.getInstance().createActuator(key,
				FooActuatorImpl.class));

		assertTrue(BonsaiManager.getInstance().createActuator(key,
				FooActuator.class) instanceof FooActuatorImpl);
		assertTrue(BonsaiManager.getInstance().createActuator(key,
				FooActuatorImpl.class) instanceof FooActuatorImpl);

		// ensure that creating an unknown key-type combination etc fails
		try {
			BonsaiManager.getInstance().createActuator(key,
					UnknownActuator.class);
			fail("Must throw an exception for an unknown key-type combination.");
		} catch (IllegalArgumentException e) {
		}

		try {
			BonsaiManager.getInstance().createActuator(key + "test",
					FooActuatorImpl.class);
			fail("Must throw an exception for an unknown key-type combination.");
		} catch (IllegalArgumentException e) {
		}

	}

	@Test
	public void correctMemoryCreation() throws Exception {

		final String key = "testKey";

		ConfiguredCoreObjectFactory f = new ConfiguredCoreObjectFactory();
		f.setFactoryClassName(TestFactory.class.getName());
		Map<String, String> options = new HashMap<String, String>();
		f.setFactroyOptions(options);
		configurator.configuredFactories.add(f);

		ConfiguredWorkingMemory memory = new ConfiguredWorkingMemory();
		memory.setKey(key);
		memory.setMemoryClassName(FooWorkingMemoryImpl.class.getName());
		memory.setFactoryClassName(TestFactory.class.getName());
		memory.setMemorySlots(new HashSet<ConfiguredMemorySlot>());
		memory.setOptions(new HashMap<String, String>());
		configurator.configuredMemories.add(memory);

		BonsaiManager.getInstance().configure("dummy", configurator);

		// check that derived instantiation is possible
		assertTrue(BonsaiManager.getInstance().canCreateWorkingMemory(key));

		// wrong key should not be possible
		assertFalse(BonsaiManager.getInstance().canCreateWorkingMemory(
				key + "bla"));

		assertNotNull(BonsaiManager.getInstance().createWorkingMemory(key));

		assertTrue(BonsaiManager.getInstance().createWorkingMemory(key) instanceof WorkingMemory);

	}
}
