package de.unibi.citec.clf.bonsai.core.configuration;

/*
 * #LICENSE%
 * BonSAI [Core|Base]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.TestUtils;
import de.unibi.citec.clf.bonsai.core.configuration.ConfigurationParser.ParseException;
import java.io.IOException;
import java.net.URI;
import org.apache.log4j.BasicConfigurator;
import static org.junit.Assert.*;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

/**
 * Test cases for the {@link XmlConfigurationParser}.
 * 
 * @author jwienke
 */
public class XmlConfigurationParserTest {

	private ConfigurationParser parser;

	@BeforeClass
	public static void staticSetUp() {
		BasicConfigurator.configure();
	}

	@Before
	public void setUp() {
		parser = new XmlConfigurationParser();
	}

	@Test(expected = IOException.class)
	public void parseFileInvalidUrl() throws Exception {
		parser.parseFile(new URI("file://dummy"));
	}

	@Test
	public void parseFileEmpty() throws Exception {
		parser.parseFile(new URI(TestUtils.makeTestFileName("empty.xml")));
		assertEquals(0, parser.getConfiguredFactories().size());
		assertEquals(0, parser.getConfiguredSensors().size());
	}

	@Test(expected = ParseException.class)
	public void parseFileNotValidating() throws Exception {
		parser.parseFile(new URI(TestUtils
				.makeTestFileName("notvalidating.xml")));
	}

	@Test
	public void parseFile() throws Exception {

		parser.parseFile(new URI(TestUtils.makeTestFileName("loader.xml")));

		// check factory parsing

		assertEquals(2, parser.getConfiguredFactories().size());

		ConfiguredCoreObjectFactory xcfFactory = findFactoryByName("de.unibi.citec.clf.bonsai.xcf.XcfFactory");
		ConfiguredCoreObjectFactory dummyFactory = findFactoryByName("de.unibi.citec.clf.bonsai.dummy.DummyFactory");

		assertNotNull(xcfFactory);
		assertNotNull(dummyFactory);

		assertEquals(1, xcfFactory.getFactroyOptions().size());
		assertTrue(xcfFactory.getFactroyOptions().containsKey(
				"maxReconnections"));
		assertEquals("5", xcfFactory.getFactroyOptions()
				.get("maxReconnections"));

		assertEquals(2, dummyFactory.getFactroyOptions().size());
		assertTrue(dummyFactory.getFactroyOptions().containsKey(
				"maxReconnections"));
		assertEquals("52",
				dummyFactory.getFactroyOptions().get("maxReconnections"));
		assertTrue(dummyFactory.getFactroyOptions().containsKey("dummyArg"));
		assertEquals("47 test text",
				dummyFactory.getFactroyOptions().get("dummyArg"));

		// check sensor parsing

		assertEquals(3, parser.getConfiguredSensors().size());

		ConfiguredSensor laserSensor = findSensorByName("LaserSensor");
		ConfiguredSensor odometrySensor = findSensorByName("OdometrySensor");
		ConfiguredSensor fooSensor = findSensorByName("FooSensor");

		assertNotNull("Config must contain a sensor with name LaserSensor",
				laserSensor);
		assertNotNull("Config must contain a sensor with name OdometrySensor",
				odometrySensor);
		assertNotNull("Config must contain a sensor with name FooSensor",
				fooSensor);

		assertEquals("de.unibi.airobots.btl.data.LaserData",
				laserSensor.getDataTypeClassName());
		assertEquals("de.unibi.citec.clf.bonsai.xcf.XcfFactory",
				laserSensor.getFactoryClassName());
		assertEquals("de.unibi.citec.clf.bonsai.xcf.PublisherSensor",
				laserSensor.getSensorClassName());
		assertEquals(1, laserSensor.getOptions().size());
		assertTrue(laserSensor.getOptions().containsKey("publisherName"));
		assertEquals("LaserData", laserSensor.getOptions().get("publisherName"));

		assertEquals("de.unibi.airobots.btl.data.OdometryData",
				odometrySensor.getDataTypeClassName());
		assertEquals("de.unibi.citec.clf.bonsai.xcf.XcfFactory",
				odometrySensor.getFactoryClassName());
		assertEquals("de.unibi.citec.clf.bonsai.xcf.PublisherSensor",
				odometrySensor.getSensorClassName());
		assertEquals(2, odometrySensor.getOptions().size());
		assertTrue(odometrySensor.getOptions().containsKey("publisherName"));
		assertEquals("OdometryData",
				odometrySensor.getOptions().get("publisherName"));
		assertTrue(odometrySensor.getOptions().containsKey("blaaa"));
		assertEquals("Foo Bar", odometrySensor.getOptions().get("blaaa"));

		assertEquals("de.unibi.airobots.btl.data.OdometryData",
				fooSensor.getDataTypeClassName());
		assertEquals("de.unibi.citec.clf.bonsai.dummy.DummyFactory",
				fooSensor.getFactoryClassName());
		assertEquals("de.unibi.citec.clf.bonsai.xcf.PublisherSensor",
				fooSensor.getSensorClassName());
		assertEquals(1, fooSensor.getOptions().size());
		assertTrue(fooSensor.getOptions().containsKey("publisherName"));
		assertEquals("DamnThing", fooSensor.getOptions().get("publisherName"));

		// check actuator parsing

		assertEquals(2, parser.getConfiguredActuators().size());

		ConfiguredActuator fooActuator = findActuatorByName("FooActuator");
		ConfiguredActuator camActuator = findActuatorByName("CamActuator");

		assertNotNull("There must be a configured FooActuator", fooActuator);
		assertNotNull("There must be a configured CamActuator", camActuator);

		assertEquals("de.unibi.citec.clf.bonsai.xcf.XcfFactory",
				fooActuator.getFactoryClassName());
		assertEquals("de.unibi.citec.clf.bonsai.actuators.FooActuator",
				fooActuator.getInterfaceClassName());
		assertEquals("de.unibi.citec.clf.bonsai.xcf.FooActuator",
				fooActuator.getActuatorClassName());
		assertEquals(2, fooActuator.getOptions().size());
		assertTrue(fooActuator.getOptions().containsKey("serverName"));
		assertEquals("FooServer", fooActuator.getOptions().get("serverName"));
		assertTrue(fooActuator.getOptions().containsKey("keepAlive"));
		assertEquals("true", fooActuator.getOptions().get("keepAlive"));

		assertEquals("de.unibi.citec.clf.bonsai.dummy.DummyFactory",
				camActuator.getFactoryClassName());
		assertEquals("de.unibi.citec.clf.bonsai.actuators.CamActuator",
				camActuator.getInterfaceClassName());
		assertEquals("de.unibi.citec.clf.bonsai.dummy.CamActuatorFoo",
				camActuator.getActuatorClassName());
		assertEquals(2, camActuator.getOptions().size());
		assertTrue(camActuator.getOptions().containsKey("serverNameBla"));
		assertEquals("CamServer", camActuator.getOptions().get("serverNameBla"));
		assertTrue(camActuator.getOptions().containsKey("keepAliveAll"));
		assertEquals("false", camActuator.getOptions().get("keepAliveAll"));

		// check memory parsing

		assertEquals(2, parser.getConfiguredWorkingMemories().size());

		ConfiguredWorkingMemory myMemory = findMemoryByName("MyMemory");
		ConfiguredWorkingMemory myMemory1 = findMemoryByName("MyMemory1");

		assertNotNull("There must be a configured MyMemory", myMemory);
		assertNotNull("There must be a configured MyMemory1", myMemory1);

		assertEquals("de.unibi.citec.clf.bonsai.core.TestFactory",
				myMemory.getFactoryClassName());
		assertEquals("de.unibi.citec.clf.bonsai.core.TestMemory",
				myMemory.getMemoryClassName());
		assertEquals(1, myMemory.getOptions().size());
		assertTrue(myMemory.getOptions().containsKey("memoryName"));
		assertEquals("LongTerm", myMemory.getOptions().get("memoryName"));

		assertEquals(1, myMemory.getMemorySlots().size());
		assertEquals("java.lang.String", myMemory.getMemorySlots().iterator()
				.next().getDataTypeClassName());
		assertEquals("de.unibi.citec.clf.bonsai.core.TestStringSlot", myMemory
				.getMemorySlots().iterator().next().getSlotClassName());
	}

	private ConfiguredSensor findSensorByName(String name) {
		for (ConfiguredSensor s : parser.getConfiguredSensors()) {
			if (s.getKey().equals(name)) {
				return s;
			}
		}
		return null;
	}

	private ConfiguredCoreObjectFactory findFactoryByName(String className) {
		for (ConfiguredCoreObjectFactory f : parser.getConfiguredFactories()) {
			if (f.getFactoryClassName().equals(className)) {
				return f;
			}
		}
		return null;
	}

	private ConfiguredWorkingMemory findMemoryByName(String key) {
		for (ConfiguredWorkingMemory f : parser.getConfiguredWorkingMemories()) {
			if (f.getKey().equals(key)) {
				return f;
			}
		}
		return null;
	}

	private ConfiguredActuator findActuatorByName(String key) {
		for (ConfiguredActuator f : parser.getConfiguredActuators()) {
			if (f.getKey().equals(key)) {
				return f;
			}
		}
		return null;
	}

	public void allowMultipleParsing() throws Exception {
		parser.parseFile(new URI(TestUtils.makeTestFileName("loader.xml")));
		parser.parseFile(new URI(TestUtils.makeTestFileName("loader.xml")));
	}

	@Test(expected = IllegalStateException.class)
	public void noSensorsAfterParsingError() throws Exception {
		try {
			parser.parseFile(new URI(TestUtils
					.makeTestFileName("notvalidating.xml")));
		} catch (Exception e) {
			// ignore this
		}
		parser.getConfiguredSensors();
	}

	@Test(expected = IllegalStateException.class)
	public void noFactoriesAfterParsingError() throws Exception {
		try {
			parser.parseFile(new URI(TestUtils
					.makeTestFileName("notvalidating.xml")));
		} catch (Exception e) {
			// ignore this
		}
		parser.getConfiguredFactories();
	}

}
