package de.unibi.citec.clf.bonsai.xcf;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.Map.Entry;

import org.apache.log4j.Logger;

import net.sf.xcf.InitializeException;
import net.sf.xcf.Subscriber;
import net.sf.xcf.XcfManager;
import net.sf.xcf.naming.NameNotFoundException;

/**
 * This class {@link SubscriberReconnectable}s based on a xcf publisher and
 * ensures reconnection if the publisher dies and is reactivated afterwards.
 * 
 * @author jwienke
 */
public class SubscriberReconnectableManager extends
        ReconnectableManager<SubscriberReconnectable> {

    private Logger logger = Logger.getLogger(getClass());

    /**
     * {@link XcfManager} used for communication.
     */
    private XcfManager xcfManager;

    /**
     * Lock object for {@link #subscribersByName} and
     * {@link #reconnectablesBySubName}.
     */
    private Object subscriberLock = new Object();

    /**
     * Map of subscribers to watch by their names. A <code>null</code> value
     * means that the subscriber could not be created and after a successful
     * creation, all {@link SubscriberReconnectable} interested in this
     * subscriber must be reinitialized.
     * 
     * Always synchronize on {@link #subscriberLock} before using this map.
     */
    private Map<String, Subscriber> subscribersByName =
            new HashMap<String, Subscriber>();

    /**
     * Managed {@link SubscriberReconnectable}s by subscriber name.
     * 
     * Always synchronize on {@link #subscriberLock} before using this list.
     */
    private Map<String, Set<SubscriberReconnectable>> reconnectablesBySubName =
            new HashMap<String, Set<SubscriberReconnectable>>();

    /**
     * Timer used to schedule the checking for subscribers.
     */
    private Timer checkTimer = new Timer("subscriber check timer");

    /**
     * The actual task executed by {@link #checkTimer} to check for subscribers.
     */
    private SubscriberChecker subscriberCheckTask = new SubscriberChecker();

    /**
     * Thread that checks all subscribers for being alive.
     * 
     * @author jwienke
     */
    private class SubscriberChecker extends TimerTask {

        @Override
        public void run() {

            logger.debug("Checking subscribers starts...");

            synchronized (subscriberLock) {

                // iterate over all managed subscriptions
                for (Entry<String, Subscriber> subscriberEntry : subscribersByName
                        .entrySet()) {

                    String subscriberName = subscriberEntry.getKey();

                    try {

                        // try to connect to this subscriber
                        Subscriber checkSubscriber =
                                xcfManager.createSubscriber(subscriberName);

                        logger.debug("Subscriber '" + subscriberName
                                + "' seems to be alive.");

                        // if connection succeeded and there was no connection
                        // before this attempt, use the new subscriber and
                        // notify reconnectables
                        if (subscriberEntry.getValue() == null) {

                            logger.info("Subscriber '" + subscriberName
                                    + "' was successfully (re-)connected. "
                                    + "Informing reconnectables about this.");

                            subscribersByName.put(subscriberName,
                                    checkSubscriber);

                            // maybe there are no reconnectables for this
                            // subscriber...
                            if (!reconnectablesBySubName
                                    .containsKey(subscriberName)) {
                                return;
                            }

                            for (SubscriberReconnectable reconnectable : reconnectablesBySubName
                                    .get(subscriberName)) {
                                reconnectable.resetSubscriber(checkSubscriber);
                            }
                        } else {
                            // checkSubscriber.deactivate();
                            xcfManager.deactivate(checkSubscriber);
                        }

                        // CHECKSTYLE:OFF catch everything here to not break the
                        // checking thread
                    } catch (Throwable e) {
                        // CHECKSTYLE:ON

                        // connection to subscriber failed, mark it as
                        // unconnected

                        logger.warn("Subscriber with name '" + subscriberName
                                + "' could not be connected "
                                + "or reinitializing on of "
                                + "the reconnectables failed. "
                                + "Marking this subscriber as unconnected.");
                        logger.debug(e);

                        subscribersByName.put(subscriberName, null);

                    }

                }

            }

            logger.debug("Checking subscribers finished");

        }

    }

    /**
     * Constructor using default check interval
     * {@link ReconnectableManager#DEFAULT_CHECK_INTERVAL_IN_MS}.
     * 
     * @param xcfManager
     *            xcf manager to use
     */
    public SubscriberReconnectableManager(XcfManager xcfManager) {
        this(xcfManager, DEFAULT_CHECK_INTERVAL_IN_MS);
    }

    /**
     * Constructor.
     * 
     * @param xcfManager
     *            xcf manager to use
     * @param checkIntervalMs
     *            interval for the subscriber check thread in ms
     */
    public SubscriberReconnectableManager(XcfManager xcfManager,
            long checkIntervalMs) {
        this.xcfManager = xcfManager;

        checkTimer.scheduleAtFixedRate(subscriberCheckTask, checkIntervalMs,
                checkIntervalMs);
    }

    /**
     * 
     * {@inheritDoc}
     */
    @Override
    public void initializeXcfTypes(Set<String> subscriberNames)
            throws XcfTypeInitializationException {

        logger.info("Initializing subscribers: " + subscriberNames);

        synchronized (subscriberLock) {

            for (String subscriberName : subscriberNames) {
                if (!subscribersByName.containsKey(subscriberName)) {
                    createSubscriber(subscriberName);
                }
            }

        }

    }

    /**
     * Creates a subscriber with the specified name regarding the specified
     * error policy set in {@link #errorOnSubscriberInitialization}. If the
     * subscriber could not be created and
     * {@link #errorOnSubscriberInitialization} is set to <code>false</code>,
     * then <code>null</code> will be inserted in {@link #subscribersByName}.
     * Otherwise a working subscriber instance is placed in this map.
     * 
     * @param subscriberName
     *            subscriber name
     * @throws XcfTypeInitializationException
     *             thrown if connection to this subscriber fails and
     *             {@link #isErrorOnXcfTypeInitialization()} is
     *             <code>true</code>
     */
    private void createSubscriber(String subscriberName)
            throws XcfTypeInitializationException {

        synchronized (subscriberLock) {

            try {

                subscribersByName.put(subscriberName,
                        xcfManager.createSubscriber(subscriberName));

            } catch (NameNotFoundException e) {

                if (isErrorOnXcfTypeInitialization()) {
                    throw new XcfTypeInitializationException(
                            "No publisher with name '" + subscriberName
                                    + "' found.", e);
                } else {
                    logger.warn("Right now no subscriber for name '"
                            + subscriberName + "' could be created. "
                            + "Initialization will be delayed.");
                    logger.debug(e);
                    subscribersByName.put(subscriberName, null);
                }

            } catch (InitializeException e) {

                if (isErrorOnXcfTypeInitialization()) {
                    throw new XcfTypeInitializationException(e);
                } else {
                    logger.warn("Right now no subscriber for name '"
                            + subscriberName + "' could be created. "
                            + "Initialization will be delayed.");
                    logger.debug(e);
                    subscribersByName.put(subscriberName, null);
                }

            }

        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void initializeAndManage(SubscriberReconnectable reconnectable)
            throws XcfTypeInitializationException {

        String subscriberName = reconnectable.getSubscriberName();

        logger.info("Got request to initialize and manage new reconnectable: "
                + reconnectable + " for subscriber with name '"
                + subscriberName + "'.");

        synchronized (subscriberLock) {

            // create new subscriber if needed
            if (!subscribersByName.containsKey(subscriberName)) {

                createSubscriber(subscriberName);

                logger.warn("There was no subscriber for name '"
                        + subscriberName
                        + "' initialized. Creation was needed.");

            }

            assert subscribersByName.containsKey(subscriberName);

            // initialize reconnectable with the subscriber
            if (subscribersByName.get(subscriberName) != null) {
                Subscriber subscriber = subscribersByName.get(subscriberName);
                logger.info("Initially setting subscriber " + subscriber
                        + " for " + reconnectable);
                reconnectable.resetSubscriber(subscriber);
            } else {
                logger.warn("There was no current connection "
                        + "to a subscriber with name '" + subscriberName
                        + "'. Initialization of " + reconnectable
                        + " postponed.");

            }

            // add reconnectable to the list of managed reconnectables
            if (!reconnectablesBySubName.containsKey(subscriberName)) {
                reconnectablesBySubName.put(subscriberName,
                        new HashSet<SubscriberReconnectable>());
            }
            reconnectablesBySubName.get(subscriberName).add(reconnectable);

        }

    }

}
