package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.memory.DocumentNotFoundException;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;

import org.apache.log4j.Logger;

import de.unibi.citec.clf.bonsai.actuators.BtlMemoryActuator;
import de.unibi.citec.clf.bonsai.core.Actuator;
import de.unibi.citec.clf.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.citec.clf.bonsai.xcf.util.XPathUtilities;
import de.unibi.airobots.btl.Type;
import de.unibi.airobots.btl.Type.NoSourceDocumentException;

/**
 * A simple {@link Actuator} for BTL based types that uses their static
 * <code>fromDocument</code> method to extract the data. This allows this
 * actuator to be used for every ActiveMemory that provides a BTL parsable
 * result.
 * 
 * @param <T>
 *            BTL type that is inserted or replaced by this actuator.
 * 
 * @author lkettenb
 */
public class XcfBtlMemoryActuator<T extends Type> implements
		BtlMemoryActuator<T>, ActiveMemoryReconnectable {

	/*
	 * Error messages.
	 */
	private static final String ERR_MEM_NOT_RECHABLE = "Scene memory is not reachable.";
	private static final String ERR_OBJ_NOT_FOUND = "Object can not be updated. "
			+ "Corresponding document not found in memory.";
	private String xPath;
	/**
	 * Instance of the active memory that is used.
	 */
	private ActiveMemory activeMemory = null;
	/**
	 * Name of the active memory.
	 */
	private String memoryName = "";
	/**
	 * Lock for synchronized memory access.
	 */
	private final Object memoryLock = new Object();
	/**
	 * The log.
	 */
	private Logger logger = Logger.getLogger(this.getClass());

	/**
	 * Constructor.
	 * 
	 * @param memoryName
	 *            Name of the active memory to work on.
	 */
	public XcfBtlMemoryActuator(String memoryName, String xpath) {
		this.memoryName = memoryName;
		this.xPath = xpath;
	}

	@Override
	public void insert(T object, Class<? extends Type> type) throws IOException {
		if (activeMemory == null) {
			logger.error(ERR_MEM_NOT_RECHABLE);
		}
		XOPData xopData;
		Document rawDoc = object.toDocument(type);

		if (xPath.equals("") || xPath.equals("/")) {
			xopData = new XOPData(rawDoc);
		} else {
			xopData = new XOPData(XPathUtilities.addParentXPath(rawDoc, xPath));
		}
		try {
			activeMemory.insert(xopData);
		} catch (MemoryException e) {
			logger.error(ERR_MEM_NOT_RECHABLE);
			logger.debug(e);
			throw new IOException(e);
		}
	}

	@Override
	public void replace(T object) throws IOException, NoSourceDocumentException {
		if (activeMemory == null) {
			logger.error(ERR_MEM_NOT_RECHABLE);
		} else {
			try {

				// TODO: ensure that memory id attribute is present!

				// copy source document
				Element source = new Element(object.getSourceDocument());

				// fill source document with new data
				object.fillInto(source);
				activeMemory.replace(new XOPData(new Document(source)));
			} catch (DocumentNotFoundException e) {
				logger.error(ERR_OBJ_NOT_FOUND);
				logger.debug(e);
				throw new IOException(e);
			} catch (MemoryException e) {
				logger.error(ERR_MEM_NOT_RECHABLE);
				logger.debug(e);
				throw new IOException(e);
			}
		}
	}

	@Override
	public String getActiveMemoryName() {
		return memoryName;
	}

	@Override
	public void resetActiveMemory(ActiveMemory activeMemory) {
		synchronized (memoryLock) {
			this.activeMemory = activeMemory;
		}
	}
}
