package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.airobots.btl.data.navigation.DriveData;
import de.unibi.citec.clf.bonsai.actuators.NavigationActuator;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.navigation.GlobalPlan;
import de.unibi.airobots.btl.data.navigation.GoalReachedData;
import de.unibi.airobots.btl.data.navigation.NavigationGoalData;
import de.unibi.airobots.btl.data.navigation.TurnData;
import de.unibi.airobots.btl.units.AngleUnit;
import de.unibi.airobots.btl.units.LengthUnit;
import de.unibi.airobots.btl.units.RotationalSpeedUnit;
import de.unibi.airobots.btl.units.SpeedUnit;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.transport.XOPData;
import nu.xom.Document;
import nu.xom.Element;
import org.apache.log4j.Logger;

/**
 * Actuator which accesses the robots navigation interface. Goals can be
 * provided using NavigationGoalData and are automatically approached using
 * obstacle avoidance.
 * 
 * @author lschilli
 * @author jwienke
 */
public class XcfNavigationActuator implements NavigationActuator,
        RemoteServerReconnectable {

    private static final String DEFAULT_GET_GOAL_METHOD = "getGoal";
    private static final String DEFAULT_DRIVE_METHOD = "drive";
    private static final String DEFAULT_TURN_METHOD = "turn";
    private static final String DEFAULT_MANUAL_STOP_METHOD = "manualStop";
    private static final String DEFAULT_GOAL_REACHED_METHOD = "goalReached";
    private static final String DEFAULT_DRIVETURN_DONE_METHOD = "driveTurnDone";
    private static final String DEFAULT_SET_GOAL_METHOD = "setGoal";
    private static final String DEFAULT_TRY_GOAL_METHOD = "tryGoal";

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName = "";
    private RemoteServer remoteServer = null;
    private Object remoteServerLock = new Object();

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server to connect to
     */
    public XcfNavigationActuator(String remoteServerName) {
        this.remoteServerName = remoteServerName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void setGoal(NavigationGoalData data) throws IOException {
        try {

            Document doc =
                    new Document(new Element(NavigationGoalData.getBaseTag()));
            data.fillInto(doc.getRootElement());

            logger.debug("Calling " + DEFAULT_SET_GOAL_METHOD + " with"
                    + doc.toXML());

            synchronized (remoteServerLock) {
                remoteServer.callMethod(DEFAULT_SET_GOAL_METHOD, new XOPData(
                        doc));
            }

            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }
    
    /**
     * {@inheritDoc}
     */
    @Override
    public GlobalPlan tryGoal(NavigationGoalData data) throws IOException {
        try {

            Document doc =
                    new Document(new Element(NavigationGoalData.getBaseTag()));
            data.fillInto(doc.getRootElement());

            logger.debug("Calling " + DEFAULT_TRY_GOAL_METHOD + " with"
                    + doc.toXML());

            XOPData xop;
            synchronized (remoteServerLock) {
                xop = remoteServer.callMethod(DEFAULT_TRY_GOAL_METHOD, new XOPData(
                        doc));
            }
            
            GlobalPlan plan = new GlobalPlan();
            GlobalPlan.fromElement(xop.getDocument().getRootElement(), plan);

            return plan;
            
            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public GoalReachedData isGoalReached() throws IOException {
        try {
            logger.debug("Calling " + DEFAULT_GOAL_REACHED_METHOD);
            XOPData xop;
            synchronized (remoteServerLock) {
                xop =
                        remoteServer.callMethod(DEFAULT_GOAL_REACHED_METHOD,
                                new XOPData());
            }
            GoalReachedData grd = new GoalReachedData();
            GoalReachedData
                    .fromElement(xop.getDocument().getRootElement(), grd);
            return grd;
            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void manualStop() throws IOException {
        try {
            logger.debug("Calling " + DEFAULT_MANUAL_STOP_METHOD);
            synchronized (remoteServerLock) {
                remoteServer.callMethod(DEFAULT_MANUAL_STOP_METHOD,
                        new XOPData());
            }
            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHEKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public NavigationGoalData getCurrentGoal() throws IOException {
        try {

            XOPData xop;
            synchronized (remoteServerLock) {
                xop =
                        remoteServer.callMethod(DEFAULT_GET_GOAL_METHOD,
                                new XOPData());
            }
            NavigationGoalData goal = new NavigationGoalData();
            NavigationGoalData.fromElement(xop.getDocument().getRootElement(),
                    goal);
            return goal;

            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }
    }

    @Override
    public void drive(double distance, LengthUnit unit, double speed, SpeedUnit sunit) throws IOException {
        
        DriveData data = new DriveData();
        data.setDistance(distance, unit);
        data.setSpeed(speed, sunit);
        
        try {

            Document doc =
                    new Document(new Element(DriveData.getBaseTag()));
            data.fillInto(doc.getRootElement());

            logger.debug("Calling " + DEFAULT_DRIVE_METHOD + " with"
                    + doc.toXML());

            synchronized (remoteServerLock) {
                remoteServer.callMethod(DEFAULT_DRIVE_METHOD, new XOPData(
                        doc));
            }

            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

    @Override
    public void turn(double angle, AngleUnit unit, double speed, RotationalSpeedUnit sunit) throws IOException {
        
        TurnData data = new TurnData();
        data.setAngle(angle, unit);
        data.setSpeed(speed, sunit);
        
        try {

            Document doc =
                    new Document(new Element(TurnData.getBaseTag()));
            data.fillInto(doc.getRootElement());

            logger.debug("Calling " + DEFAULT_TURN_METHOD + " with"
                    + doc.toXML());

            synchronized (remoteServerLock) {
                remoteServer.callMethod(DEFAULT_TURN_METHOD, new XOPData(
                        doc));
            }

            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
        
    }

    @Override
    public GoalReachedData driveTurnDone() throws IOException {
        try {
            logger.debug("Calling " + DEFAULT_DRIVETURN_DONE_METHOD);
            XOPData xop;
            synchronized (remoteServerLock) {
                xop =
                        remoteServer.callMethod(DEFAULT_DRIVETURN_DONE_METHOD,
                                new XOPData());
            }
            GoalReachedData grd = new GoalReachedData();
            GoalReachedData
                    .fromElement(xop.getDocument().getRootElement(), grd);
            return grd;
            // CHECKSTYLE:OFF every error here is an error communicating with
            // the server. Either a connection problem or a semantic one
        } catch (Exception e) {
            // CHECKSTYLE:ON
            throw new IOException(e);
        }
    }

}
