package de.unibi.citec.clf.bonsai.xcf.actuators;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.actuators.SwissRangerControl;
import de.unibi.citec.clf.bonsai.xcf.RemoteServerReconnectable;
import de.unibi.airobots.btl.data.grasp.PoseData;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import net.sf.xcf.MethodNotFoundException;
import net.sf.xcf.RemoteServer;
import net.sf.xcf.UserException;
import net.sf.xcf.transport.XOPData;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import nu.xom.ParsingException;
import org.apache.log4j.Logger;

/**
 * This class allows you to (de-) activate the virtual laser.
 * 
 * @author jotto
 * @author jwienke
 * @author lziegler
 */
public class XcfSwissRangerControl implements RemoteServerReconnectable,
        SwissRangerControl {

    private Logger logger = Logger.getLogger(getClass());

    private String remoteServerName;
    private Object remoteServerLock = new Object();
    private RemoteServer remoteServer;

    /**
     * Constructor.
     * 
     * @param remoteServerName
     *            name of the remote server used to control the swiss ranger
     */
    public XcfSwissRangerControl(String remoteServerName) {
        this.remoteServerName = remoteServerName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void deactivateVirtualLaser() throws IOException {

        Element elem = new Element("deactive");
        Document d = new Document(elem);

        try {
            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("deactivate", new XOPData(
                                d));
            }
            logger.debug(res.getDocument().toString());
        } catch (MethodNotFoundException e) {
            logger.debug("the method is not registered on the server", e);
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void activateVirtualLaser() throws IOException {

        Element elem = new Element("activate");
        Document d = new Document(elem);

        try {
            XOPData res;
            synchronized (remoteServerLock) {
                res = this.remoteServer.callMethod("activate", new XOPData(d));
            }
            logger.debug(res.getDocument());
        } catch (MethodNotFoundException e) {
            logger.debug("the method is not registered on the server", e);
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        }

    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Map<String, ErrorHandlingPolicy> getRemoteServerNames() {
        Map<String, ErrorHandlingPolicy> servers =
                new HashMap<String, ErrorHandlingPolicy>();
        servers.put(remoteServerName, ErrorHandlingPolicy.SELF_HEALING);
        return servers;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetRemoteServer(RemoteServer remoteServer) {
        synchronized (remoteServerLock) {
            this.remoteServer = remoteServer;
        }

    }

    @Override
    public PoseData getRawPosition(int x, int y) throws IOException {
        return getRawPosition(x, y, 1, 1);
    }

    @Override
    public PoseData getRawPosition(int x, int y, int radius) throws IOException {
        return getRawPosition(x, y, radius, 1);
    }

    @Override
    public PoseData getRawPosition(int x, int y, int radius, int frames)
            throws IOException {
        
        if (radius < 1 || frames < 1) {
            throw new IOException("radius or frames not in range!");
        }
        
        Element elem = new Element("getRawPosition");
        elem.addAttribute(new Attribute("x", Integer.toString(x)));
        elem.addAttribute(new Attribute("y", Integer.toString(y)));
        elem.addAttribute(new Attribute("radius", Integer.toString(radius)));
        elem.addAttribute(new Attribute("frames", Integer.toString(frames)));
        Document d = new Document(elem);
        PoseData pose;

        try {
            XOPData res;
            synchronized (remoteServerLock) {
                res =
                        this.remoteServer.callMethod("getRawPosition",
                                new XOPData(d));
            }
            logger.debug(res.getDocument().toString());
            pose = PoseData.fromDocument(res.getDocument(), PoseData.class);
        } catch (MethodNotFoundException e) {
            logger.debug("the method is not registered on the server", e);
            throw new IOException(e);
        } catch (UserException e) {
            throw new IOException(e);
        } catch (ParsingException e) {
            throw new IOException(e);
        }
        return pose;
    }

}
