package de.unibi.citec.clf.bonsai.xcf.configuration;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.SensorToConfigure;
import de.unibi.citec.clf.bonsai.core.Sensor;
import de.unibi.citec.clf.bonsai.xcf.sensors.XcfBtlMemorySensor;
import de.unibi.airobots.btl.List;
import de.unibi.airobots.btl.Type;
import java.util.Map;

/**
 * Configuration for {@link XcfBtlMemorySensor}.
 * <p>
 * Available options:
 * <ul>
 * <li>{@link ConfiguredXcfSensor#OPTION_PUBLISHER_NAME} (required): name of the
 * publisher to register on</li>
 * </ul>
 * 
 * @author lziegler
 * @param <T>
 *            sensor data type
 */
public class ConfiguredXcfBtlMemorySensor<T extends Type> extends
		ConfiguredXcfSensorListable<List<T>, T> {
	
	/**
	 * This method is a helper method to create {@link Class} objects for the
	 * generic {@link List} type.
	 * 
	 * @param type The type of the {@link List}'s elements.
	 * @return A {@link Class} object representing the generic {@link List}.
	 */
	@SuppressWarnings("unchecked")
	public static <T extends Type> Class<List<T>> createListClassOfType(
			Class<T> type) {
		return (Class<List<T>>) new List<T>(type).getClass();
	}

	public static final String OPTION_XPATH = "xPath";

	private String xPath = "";

	/**
	 * Option for publisher based sensors defining the memory name.
	 */
	public static final String OPTION_MEMORY_NAME = "memoryName";

	/**
	 * Data type to be returned by the sensor.
	 */
	private Class<T> dataType = null;

	/**
	 * Name of the configured memory if {@link #parseOptions(Map)} was used.
	 */
	private String memoryName = "";

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String[] getActiveMemoriesToInitialize() {
		return new String[] { memoryName };
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public String[] getSubscribersToInitialize() {
		return new String[] {};
	}

	/**
	 * Sets the data type the configured sensor is intended to return.
	 * 
	 * @param dataType
	 *            new data type
	 */
	@Override
	public void setDataType(Class<T> dataType) {
		this.dataType = dataType;
	}

	/**
	 * Returns the data type the configured sensor will return.
	 * 
	 * @return data type class
	 */
	@Override
	public Class<T> getDataType() {
		return dataType;
	}

	/**
	 * Returns the configured memory name for the sensor if
	 * {@link #parseOptions(Map)} was used.
	 * 
	 * @return publisher name
	 */
	public String getMemoryName() {
		return memoryName;
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public Sensor<T> createInstance() {
		return new XcfBtlMemorySensor<T>(getDataType(), getMemoryName(), xPath);
	}

	@Override
	public Sensor<List<T>> createInstanceListable() {
		return new XcfBtlMemorySensor<List<T>>(
				createListClassOfType(getDataType()), getMemoryName(), xPath);
	}

	@Override
	public void parseOptions(Map<String, String> options)
			throws IllegalArgumentException {

		if (!options.containsKey(OPTION_MEMORY_NAME)) {
			throw new IllegalArgumentException(
					"Memory name missing in arguments.");
		}

		memoryName = options.get(OPTION_MEMORY_NAME);

		if (options.containsKey(OPTION_XPATH)) {
			this.xPath = options.get(OPTION_XPATH);
		}
	}

	/**
	 * Checks if the sensor that can be produced is suitable for the requested
	 * configuration.
	 * 
	 * @param sensor
	 *            requested configuration
	 * @return <code>true</code> if suitable
	 */
	public static boolean isSuitableFor(SensorToConfigure sensor) {
		return sensor.getSensorClass().equals(XcfBtlMemorySensor.class)
				&& Type.class.isAssignableFrom(sensor.getDataTypeClass());
	}
}
