package de.unibi.citec.clf.bonsai.xcf.configuration;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.CoreObjectFactory.SensorToConfigure;
import de.unibi.citec.clf.bonsai.core.Sensor;
import java.util.Map;

/**
 * Base class for XCF-based sensor configuration objects. The default
 * implementation of {@link #parseOptions(Map)} requires a publisher name for
 * convenience reasons.
 * 
 * @author jwienke
 * @param <T>
 *            data type returned by the sensor
 */
public abstract class ConfiguredXcfSensor<T> extends ConfiguredXcfCoreObject {

    /**
     * Option for publisher based sensors defining the publisher name.
     */
    public static final String OPTION_PUBLISHER_NAME = "publisherName";

    /**
     * Data type to be returned by the sensor.
     */
    private Class<T> dataType = null;

    /**
     * Name of the configured publisher if {@link #parseOptions(Map)} was used.
     */
    private String publisherName = "";

    /**
     * Create a new specific sensor instance.
     * 
     * @return new instance
     */
    public abstract Sensor<T> createInstance();

    /**
     * {@inheritDoc}
     * 
     * Default implementation parsing the publisher name.
     */
    @Override
    public void parseOptions(Map<String, String> options)
            throws IllegalArgumentException {

        if (!options.containsKey(OPTION_PUBLISHER_NAME)) {
            throw new IllegalArgumentException(
                    "Publisher name missing in arguments.");
        }

        publisherName = options.get(OPTION_PUBLISHER_NAME);

    }

    /**
     * Override this and check for suitable data type!
     * 
     * @param sensor
     *            requested sensor
     * @return <code>true</code> if configuration can create such a requested
     *         sensor
     */
    public static boolean isSuitableFor(SensorToConfigure sensor) {
        return false;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String[] getSubscribersToInitialize() {
        return new String[] { publisherName };
    }

    /**
     * Sets the data type the configured sensor is intended to return.
     * 
     * @param dataType
     *            new data type
     */
    public void setDataType(Class<T> dataType) {
        this.dataType = dataType;
    }

    /**
     * Returns the data type the configured sensor will return.
     * 
     * @return data type class
     */
    public Class<T> getDataType() {
        return dataType;
    }

    /**
     * Returns the configured publisher name for the sensor if
     * {@link #parseOptions(Map)} was used.
     * 
     * @return publisher name
     */
    public String getPublisherName() {
        return publisherName;
    }

}
