package de.unibi.citec.clf.bonsai.xcf.sensors;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import de.unibi.citec.clf.bonsai.core.SensorListener;
import de.unibi.citec.clf.bonsai.sensors.SceneMultiGridMapSensor;
import de.unibi.citec.clf.bonsai.sensors.data.DynamicGridMapList;
import de.unibi.citec.clf.bonsai.util.DataConverter;
import de.unibi.citec.clf.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.map.DynamicGridMap;
import java.io.IOException;
import java.util.*;
import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.Attachment;
import net.sf.xcf.transport.AttachmentNotFoundException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Attribute;
import nu.xom.Element;
import nu.xom.ParsingException;
import org.apache.log4j.Logger;

/**
 * A sensor to receive gridmaps from an active memory.
 * 
 * @author lziegler
 */
public class XcfSceneMultiGridMapSensor implements ActiveMemoryReconnectable,
        SceneMultiGridMapSensor {

    private Logger logger = Logger.getLogger(getClass());

    private String memoryName = "";
    private Object memoryLock = new Object();
    private Object sharedDataLock = new Object();
    private ActiveMemory memory = null;

    /**
     * Namespace of active memory xml documents.
     */
    public static final String NAMESPACE_DBXML =
            "http://www.sleepycat.com/2002/dbxml";

    private List<SensorListener<DynamicGridMapList>> listeners =
            new LinkedList<SensorListener<DynamicGridMapList>>();

    private Map<Integer, DynamicGridMap> maps =
            new HashMap<Integer, DynamicGridMap>();

    /**
     * Own listener used to intercept the memory for person changes.
     */
    private MemoryEventListener selfListener;

    /**
     * Constructor.
     * 
     * @param memoryName
     *            name of the active memory to get the persons from
     */
    public XcfSceneMultiGridMapSensor(String memoryName) {

        this.memoryName = memoryName;

        logger.debug("Create scene memory sensor for memoryName " + memoryName);

        selfListener =
                new MemoryEventAdapter(MemoryAction.ALL, new XPath(
                        DynamicGridMap.getBaseTag())) {

                    @Override
                    public void handleEvent(MemoryEvent e) {
                        handleMemoryEvent(e);
                    }
                };
    }

    private void handleMemoryEvent(MemoryEvent e) {
        try {

            synchronized (sharedDataLock) {

                // create BTL type
                XOPData data = e.getData();
                Element root = data.getDocument().getRootElement();
                if (root == null) {
                    logger.error("Received Document has no root node");
                    return;
                }
                DynamicGridMap map = new DynamicGridMap();
                DynamicGridMap.fromElement(root, map);

                // get attachment
                try {
                    if (e.getType() != MemoryAction.REMOVE) {
                        XOPData mapData = memory.getAttachment(data);
                        Attachment mapPacket =
                                mapData.getAttachment(map.getUri());
                        byte[] bytes = mapPacket.getValue();
                        map.setGridMap(DataConverter.toFloatArray(bytes));
                    }
                } catch (MemoryException ex) {
                    logger.error("Error decoding map attachment.");
                    logger.debug(ex);
                    if (ex.getCause() != null) {
                        logger.debug(ex.getCause());
                    }
                    logger.debug("Tried to get attachment for document: "
                            + data.getDocumentAsText());
                }

                // extract xcf id
                Attribute att = root.getAttribute("id", NAMESPACE_DBXML);
                if (att == null) {
                    logger.error("Received Document has no xcf id");
                    return;
                }
                int id = Integer.parseInt(att.getValue());

                MemoryAction a = e.getType();
                if (a == MemoryAction.INSERT) {
                    logger.debug("insert id " + id);
                    maps.put(id, map);
                } else if (e.getType() == MemoryAction.REPLACE) {
                    logger.debug("replace id " + id);
                    maps.put(id, map);
                } else if (e.getType() == MemoryAction.REMOVE) {
                    logger.debug("remove id " + id);
                    maps.remove(id);
                }

                for (SensorListener<DynamicGridMapList> l : listeners) {
                    l.newDataAvailable(makeList(maps.values()));
                }

            }

        } catch (ParsingException ex) {
            logger.error(ex.getMessage());
            logger.debug(ex);
        } catch (NumberFormatException ex) {
            logger.error("Received Document has not a valid xcf id");
            logger.debug(ex);
        } catch (AttachmentNotFoundException ex) {
            logger.error(ex.getMessage());
            logger.debug(ex);
            throw ex;
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public String getActiveMemoryName() {
        return memoryName;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {

        synchronized (memoryLock) {
            this.memory = activeMemory;

            try {
                memory.addListener(selfListener);
            } catch (MemoryException e) {

            }
        }
    }

    @Override
    public boolean hasNext() {
        return maps.size() > 0;
    }

    private DynamicGridMapList makeList(Collection<DynamicGridMap> o) {
        DynamicGridMapList list = new DynamicGridMapList();
        list.addAll(o);
        return list;
    }

    @Override
    public Class<DynamicGridMapList> getDataType() {
        return DynamicGridMapList.class;
    }

    @Override
    public DynamicGridMapList readLast() throws IOException,
            InterruptedException {
        return makeList(maps.values());
    }

    @Override
    public DynamicGridMapList readLast(long timeout) throws IOException,
            InterruptedException {
        return makeList(maps.values());
    }

    @Override
    public void addSensorListener(SensorListener<DynamicGridMapList> listener) {
        listeners.add(listener);
    }

    @Override
    public void removeSensorListener(SensorListener<DynamicGridMapList> listener) {
        listeners.remove(listener);
    }

}
