package de.unibi.citec.clf.bonsai.xcf.sensors;

/*
 * #LICENSE%
 * BonSAI [Core|XCF]
 * ---
 * Copyright (C) 2009 - 2013 Frederic Siepmann
 * ---
 * This file is part of Bielefeld Sensor Actuator Interface (BonSAI).
 * 
 * http://opensource.cit-ec.de/projects/bonsai
 * 
 * This file may be licensed under the terms of of the
 * GNU Lesser General Public License Version 3 (the ``LGPL''),
 * or (at your option) any later version.
 * 
 * Software distributed under the License is distributed
 * on an ``AS IS'' basis, WITHOUT WARRANTY OF ANY KIND, either
 * express or implied. See the LGPL for the specific language
 * governing rights and limitations.
 * 
 * You should have received a copy of the LGPL along with this
 * program. If not, go to http://www.gnu.org/licenses/lgpl.html
 * or write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * The development of this software was supported by the
 * Excellence Cluster EXC 277 Cognitive Interaction Technology.
 * The Excellence Cluster EXC 277 is a grant of the Deutsche
 * Forschungsgemeinschaft (DFG) in the context of the German
 * Excellence Initiative.
 * 
 * Contributors: Florian Lier, Frederic Siepmann, Leon Ziegler, Matthias 
 * Schoepfer, Adriana-Victoria Dreyer, Agnes Swadzba, Andreas Kipp, Birte 
 * Carlmeyer, Christian Ascheberg, Daniel Nacke, Dennis Wigand, Günes Minareci, 
 * Hendrik ter Horst, Ingo Killmann, Jan Pöppel, Lukas Kettenbach, Michael 
 * Zeunert, Patrick Renner, Philipp Dresselhaus, Sebastian Meyer zu Borgsen, 
 * Soufian Jebbara, Tobias Röhlig, Torben Toeniges, Viktor Losing, Viktor Richter
 * %LICENSE#
 */

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.LinkedList;
import java.util.List;

import org.apache.log4j.Logger;

import net.sf.xcf.ActiveMemory;
import net.sf.xcf.event.MemoryEvent;
import net.sf.xcf.event.MemoryEventAdapter;
import net.sf.xcf.event.MemoryEventListener;
import net.sf.xcf.memory.MemoryAction;
import net.sf.xcf.memory.MemoryException;
import net.sf.xcf.transport.XOPData;
import net.sf.xcf.xml.XPath;
import nu.xom.Attribute;
import nu.xom.Document;
import nu.xom.Element;
import de.unibi.citec.clf.bonsai.core.Sensor;
import de.unibi.citec.clf.bonsai.core.SensorListener;
import de.unibi.citec.clf.bonsai.xcf.ActiveMemoryReconnectable;
import de.unibi.airobots.btl.data.taskpattern.TaskPatternBase;


/**
 * A simple {@link Sensor} for BTL based types that uses their static
 * <code>fromDocument</code> method to extract the data. This allows this sensor
 * to be used for every XCF-TaskSensorPattern that provides a BTL parseable result.
 * @author arothert, vrichter, bcarlmey
 */
public class XcfTaskPatternSensor<T extends TaskPatternBase> implements ActiveMemoryReconnectable,
                                    Sensor<T>{

    private Logger logger = Logger.getLogger(getClass());

    private String memoryName = "";
    private Object memoryLock = new Object();
    private ActiveMemory activeMemory = null;
    private final String taskName;
    private Class<T> typeClass;
    T taskPattern = null;

    private Method fromDocumentMethod = null;
    
    private Document taskRep;     
    /**
     * Xml namespace for dbxml.
     */
    public static final String NAMESPACE_DBXML =
            "http://www.sleepycat.com/2002/dbxml";
    
    private List<SensorListener<T>> listeners =
        new LinkedList<SensorListener<T>>();
    
    /**
     * Own listener used to intercept the memory for changes.
     */
    private MemoryEventListener selfListener;
    
    /**
     * contsructor.
     * @param typeClass Type of TaskPattern
     * @param memoryName memory name
     * @param taskName base tag of the TaskPattern btl type.
     */
    public XcfTaskPatternSensor(Class<T> typeClass, String memoryName, String taskName){
        this.taskName = taskName;
        this.memoryName = memoryName;

        this.typeClass = typeClass;
        
        // find fromDocument method for the btl type
        try {
            fromDocumentMethod =
                    typeClass.getMethod("fromDocument", new Class[] {
                            Document.class, Class.class });
        } catch (SecurityException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        } catch (NoSuchMethodException e) {
            throw new IllegalArgumentException(
                    "Error finding a 'fromDocument' method in the given type "
                            + typeClass + ".", e);
        }
        
        logger.debug("Create memory sensor for memoryName " + memoryName + " "+ taskName);

        selfListener = new MemoryEventAdapter(MemoryAction.ALL, new XPath(taskName)) {

                    @Override
                    public void handleEvent(MemoryEvent e) {
                        logger.debug("inside handleMemoryEvent");
                        handleMemoryEvent(e);
                    }
                };
        //activeMemory.addListener(selfListener);
    }
    
    /**
     * handles the memory event.
     * @param e Memory Event.
     */
    protected void handleMemoryEvent(MemoryEvent e) {
        try {

            synchronized (memoryLock) {
                logger.debug("inside handleMemoryEvent");
                MemoryAction a = e.getType();
                logger.debug("MemoryAction "+ a);
                if (a != MemoryAction.INSERT &&
                    a != MemoryAction.REPLACE &&
                    a != MemoryAction.REMOVE) return;

                try {
                    taskPattern = safeExtractType(e.getData());
                    
                } catch (IOException e1) {
                    logger.error(e1);
                }
                logger.info("taskPattern "+taskPattern);
                // create BTL type
                XOPData data = e.getData();               
                Element root = data.getDocument().getRootElement();
                setTaskRep(data.getDocument());
                
                if (root == null) {
                    logger.error("Received Document has no root node");
                    return;
                }

                // extract xcf id
                Attribute att = root.getAttribute("id", NAMESPACE_DBXML);
                if (att == null) {
                    logger.error("Received Document has no xcf id");
                    return;
                }
                int id = Integer.parseInt(att.getValue());
                
                synchronized (memoryLock) {
                
                    a = e.getType();
                    if (a == MemoryAction.INSERT) {
                        logger.debug("insert id " + id);
                        this.taskPattern = taskPattern;
                    } else if (a == MemoryAction.REPLACE) {
                        logger.debug("replace id " + id);
                        this.taskPattern = taskPattern;
                    } else if (a == MemoryAction.REMOVE) {
                        logger.debug("remove id " + id);
                        this.taskPattern = null;
                    }
                    for (SensorListener<T> l : listeners) {
                        l.newDataAvailable(this.taskPattern);
                    }
                
                }

            }
        } catch (NumberFormatException ex) {
            logger.error("Received Document has not a valid xcf id");
            logger.debug(ex);
        }
        
    }

    private T safeExtractType(XOPData data) throws IOException {
        Document doc = data.getDocument();
        if (doc == null) {
            throw new IOException("Document empty in XOPData: " + data);
        }
        logger.debug("safeExtractType "+doc.toXML() + "\n" +typeClass);
        try {
            
            Object rawResult = fromDocumentMethod.invoke(null,doc, typeClass);
            if (rawResult == null) {
                throw new IOException(
                        "fromDocument method has wrong signature "
                                + "or simply returned null.");
            }
            return (T) rawResult;
        } catch (IllegalArgumentException e) {
            throw new IOException(e);
        } catch (ExceptionInInitializerError e) {
            throw new IOException(e);
        
        } catch (IllegalAccessException e) {
            throw new IOException(e);
        } catch (InvocationTargetException e) {
            throw new IOException(e);
        } 
    }

    @Override
    public String getActiveMemoryName() {
        return this.memoryName;
    }

    @Override
    public void resetActiveMemory(ActiveMemory activeMemory) {
        synchronized (memoryLock) {
            this.activeMemory = activeMemory;

            try {
                activeMemory.addListener(selfListener);
            } catch (MemoryException e) {

            }
        }         
    }


    @Override
    public T readLast() throws IOException, InterruptedException {
        return this.taskPattern;
    }

    @Override
    public T readLast(long timeout) throws IOException, InterruptedException {
        return this.taskPattern;
    }

    @Override
    public boolean hasNext() {
        return this.taskPattern != null;
    }

    @Override
    public void addSensorListener(SensorListener<T> listener) {
        listeners.add(listener);
    }

    @Override
    public void removeSensorListener(SensorListener<T> listener) {
        listeners.remove(listener);
    }
    
   
    @Override
    public Class<T> getDataType() {
        return typeClass;
    }

    public void setTaskRep(Document taskRep) {
        this.taskRep = taskRep;
    }

    public Document getTaskRep() {
        return taskRep;
    }

}
